#!/usr/bin/env python3
"""Matplotlib themes as dictionaries."""

from typing import List, Tuple
from .colors import LIGHT_COLORS, DARK_COLORS, LIGHT_SHADES, DARK_SHADES
from .sequential_palettes import SEQUENTIAL

# Colors are taken from BlueprintJS. Background-foreground colors are default in Elastic UI.

rho = {
    # text sizes
    "font.size": 12,
    "axes.labelsize": "large",
    "axes.titlesize": "x-large",
    "figure.titlesize": "xx-large",
    # text weights
    "axes.titleweight": 400,
    "figure.titleweight": 700,
    # if grid is turned on, have it go below plots
    "axes.axisbelow": "false",
    "grid.linestyle": "-",
    "grid.linewidth": 1.0,
    # turn off grid by default
    "axes.grid": "false",
    # turn off ticks on axes
    "xtick.major.size": 0,
    "xtick.minor.size": 0,
    "ytick.major.size": 0,
    "ytick.minor.size": 0,
    # sans-serif by default
    "font.family": "sans-serif",
    # use Computer Modern for LaTeX
    "mathtext.fontset": "cm",
    # turn off right and top spines
    "axes.spines.right": "false",
    "axes.spines.top": "false",
    # set default figure size to a bit more than 6.4 x 4.8
    "figure.figsize": (8, 6),
    # slightly thicker lines by default are easier to see
    "lines.linewidth": 3,
    "image.cmap": "rho_viridia",
}


rho_light = rho.copy()
rho_dark = rho.copy()

for rc, shades in [(rho_light, LIGHT_SHADES), (rho_dark, DARK_SHADES)]:
    empty, lightest, light, medium, dark, darkest = [x[1:] for x in shades]
    rc["figure.facecolor"] = empty
    rc["savefig.facecolor"] = empty
    rc["axes.facecolor"] = empty

    rc["axes.edgecolor"] = light
    rc["grid.color"] = light
    rc["figure.edgecolor"] = light
    rc["savefig.edgecolor"] = light
    rc["legend.edgecolor"] = light
    rc["boxplot.flierprops.color"] = light
    rc["boxplot.boxprops.color"] = light
    rc["boxplot.whiskerprops.color"] = light
    rc["boxplot.capprops.color"] = light

    rc["xtick.color"] = medium
    rc["ytick.color"] = medium

    rc["xtick.labelcolor"] = dark
    rc["ytick.labelcolor"] = dark
    rc["legend.labelcolor"] = dark

    rc["text.color"] = darkest
    rc["axes.labelcolor"] = darkest
    rc["axes.titlecolor"] = darkest


def boxstyle(is_dark=None) -> dict:
    """Returns kwargs that style Matplotlib/Seaborn boxplots to match the theme.
    If filled, then main color is used to fill the box. Otherwise, the box has a border
    but no fill.

    If is_dark is True/False, use that for theming. Otherwise, uses current theme."""
    if is_dark is None:
        import matplotlib.pyplot as plt

        bg = plt.rcParams["axes.facecolor"]
        fg = plt.rcParams["xtick.labelcolor"]
    else:
        empty, _lightest, _light, _medium, dark, _darkest = (
            DARK_SHADES if rho_dark else LIGHT_SHADES
        )
        bg = empty
        fg = dark

    lw = 1.5
    opts = dict(
        medianprops={"color": bg, "lw": lw},
        boxprops={"ec": bg, "lw": lw},
        flierprops={"mfc": fg},
        whiskerprops={"color": fg, "lw": lw},
        capprops={"color": fg, "lw": lw},
    )

    return opts


def unfill_boxplot(ax=None) -> None:
    """Inverts boxplot colors so the borders are the color being plotted and the inner box is empty. Modifies an axis in-place.
    If no axis is given, defaults to current axis."""

    if ax is None:
        import matplotlib.pyplot as plt

        ax = plt.gca()

    for patch in ax.patches:
        patch.set(ec=patch.get_fc())
        patch.set(fill=False)

    for i in range(len(ax.lines) // 6):
        (whis1, whis2, cap1, cap2, median, dunno) = ax.lines[i * 6 : i * 6 + 6]
        for l in (whis1, whis2, cap1, cap2, median):
            l.set_color(ax.patches[i].get_ec())


def setup(is_dark: bool, setup=True) -> Tuple[dict, List[str]]:
    """Sets up Matplotlib according to the given color scheme and pyplot module. Returns the theme and colors as a tuple, setting the theme and colormaps."""
    theme = rho_dark if is_dark else rho_light
    if setup:
        import matplotlib as mpl
        import matplotlib.pyplot as plt
        from matplotlib.colors import LinearSegmentedColormap

        # this annoyingly requires a matplotlib object, so delay until
        # we're confident that matplotlib is installed
        rho_light["axes.prop_cycle"] = mpl.cycler(
            # remove # from color beginning
            color=[x[1:] for x in LIGHT_COLORS]
        )

        rho_dark["axes.prop_cycle"] = mpl.cycler(
            # remove # from color beginning
            color=[x[1:] for x in DARK_COLORS]
        )

        for name, palette in SEQUENTIAL.items():
            cmap = palette.as_mpl_cmap()
            try:
                plt.get_cmap("rho_" + name)
            except ValueError:
                plt.register_cmap(name="rho_" + name, cmap=cmap)

            try:
                plt.get_cmap("rho_" + name + "_r")
            except ValueError:
                plt.register_cmap(name="rho_" + name + "_r", cmap=cmap.reversed())

        # https://github.com/jupyter/notebook/issues/3691

        # Some IPython versions have a bug where Matplotlib styles don't take
        # effect until the plt module is used, not just when it first imports.
        # Clearing the figure is the least intrusive way I can find to trigger
        # this update so the theme will take effect without having to put this
        # in a separate cell from the import.
        plt.clf()

        plt.style.use(theme)

    return (theme, [x["color"] for x in theme["axes.prop_cycle"]])



from .matplotlib import setup as mpl_setup
from .matplotlib import boxstyle as boxstyle
from .matplotlib import unfill_boxplot as unfill_boxplot
from .matplotlib_tweaks import smart_ticks, line_labels, ylabel_top

try:
    import altair
    from .vega import setup as vega_setup
    from .vega import RHO_LIGHT as vega_rho_light
    from .vega import RHO_DARK as vega_rho_dark
except ModuleNotFoundError:
    pass

from .sequential_palettes import *



#!/usr/bin/env python3
"""Reads a core set of sequential color scales."""

from importlib.resources import open_text
import json
from typing import Mapping
from .palettes import SequentialPalette

SEQUENTIAL_DATA = json.load(open_text("rho_plus.data", "sequential_palettes.json"))
SEQUENTIAL: Mapping[str, SequentialPalette] = {}

for name, colors in SEQUENTIAL_DATA.items():
    globals()[name] = SequentialPalette(name, colors)
    SEQUENTIAL[name] = SequentialPalette(name, colors)



"""Matplotlib tweaks to plots that enable more complicated customizations from Matplotlib."""

import textwrap

import numpy as np

import matplotlib as mpl
import matplotlib.pyplot as plt


def remove_crowded(ax=None):
    """Given an axis, removes the second and penultimate ticks if
    the distance between the final ticks and those are less than half of the
    average distance."""
    if ax is None:
        ax = plt.gca()

    # check in view coordinates to adjust for log axes and the like
    ticks = [ax.get_xticks(), ax.get_yticks()]
    xticklocs = ax.transData.transform(
        np.vstack([ticks[0], np.ones_like(ticks[0]) * ticks[1][0]]).T
    )[:, 0]
    yticklocs = ax.transData.transform(
        np.vstack([np.ones_like(ticks[1]) * ticks[0][0], ticks[1]]).T
    )[:, 1]

    tick_locs_list = [xticklocs, yticklocs]
    for i, tick_locs in enumerate(tick_locs_list):
        meandist = np.diff(tick_locs).mean()
        if tick_locs[1] - tick_locs[0] < 0.5 * meandist:
            ticks[i] = np.array([ticks[i][0], *ticks[i][2:]])

        if tick_locs[-1] - tick_locs[-2] < 0.5 * meandist:
            ticks[i] = np.array([*ticks[i][:-2], ticks[i][-1]])

    ax.set_xticks(ticks[0])
    ax.set_yticks(ticks[1])


def smart_ticks(ax=None, xaxis=True, yaxis=True):
    """Sets the axis limits so they start and end at the minimum and maximum for each axis, a la Tufte."""
    if ax is None:
        ax = plt.gca()

    bboxes = []
    for child in ax.get_children():
        if isinstance(child, mpl.collections.Collection):
            bbox = child.get_datalim(ax.transData)
        elif isinstance(child, mpl.lines.Line2D):
            xdata, ydata = child.get_data()
            if len(xdata):
                bbox = mpl.transforms.Bbox.from_extents(
                    xdata.min(), ydata.min(), xdata.max(), ydata.max()
                )
            else:
                continue
        else:
            continue
        bboxes.append(bbox)

    if not len(bboxes):
        return None

    bbox = mpl.transforms.Bbox.union(bboxes)
    ax.spines["bottom"].set_bounds(bbox.xmin, bbox.xmax)
    ax.spines["left"].set_bounds(bbox.ymin, bbox.ymax)

    # ignore fixed locators, which are often not numeric
    if xaxis and not isinstance(ax.xaxis.get_major_locator(), mpl.ticker.FixedLocator):
        xlabels = ax.xaxis.get_major_formatter().format_ticks([bbox.xmin, bbox.xmax])
        xlabels = [
            xlabels[0],
            *ax.xaxis.get_major_formatter().format_ticks(ax.get_xticks()[1:-1]),
            xlabels[-1],
        ]
        xticks = [bbox.xmin, *ax.get_xticks()[1:-1], bbox.xmax]
        ax.set_xticks(xticks, xlabels)

    if yaxis and not isinstance(ax.yaxis.get_major_locator(), mpl.ticker.FixedLocator):
        ylabels = ax.yaxis.get_major_formatter().format_ticks([bbox.ymin, bbox.ymax])
        ylabels = [
            ylabels[0],
            *ax.yaxis.get_major_formatter().format_ticks(ax.get_yticks()[1:-1]),
            ylabels[-1],
        ]
        yticks = [bbox.ymin, *ax.get_yticks()[1:-1], bbox.ymax]
        ax.set_yticks(yticks, ylabels)

    remove_crowded(ax)


def spread(x, dmin):
    """Returns a shifted x' with minimum total shift such that no two points in x' are within dmin.
    If dmin is array-like, it must be the same shape as x, and notes the minimum space each element
    must have on either side."""
    x = np.array(x, copy=False)
    if len(x) <= 1:
        return x

    # we need to work on sorted x, but we also need to remember the indices so we can adjust the right ones
    sort_inds = np.argsort(x)
    x_sort = x.copy()[sort_inds]

    dmin = np.broadcast_to(dmin, x.shape)[sort_inds]

    # Algorithms for Minimizing the Movements of Spreading Points in Linear Domains
    # Li, Wang 2015
    # https://cccg.ca/proceedings/2015/08.pdf

    def adjust_rightward(x_sort):
        """Returns necessary adjustments if only addition is allowed."""
        adj = np.zeros_like(x_sort, dtype=np.float64)
        for i in range(1, len(x_sort)):
            # previous element's new location is x_sort[i-1] + adj[i-1]
            # move to max(x_sort[i], prev + dmin)

            # dmin = max(dmin[i], dmin[i-1])

            local_dmin = max(dmin[i], dmin[i - 1])
            prev = x_sort[i - 1] + adj[i - 1]
            new_x = max(x_sort[i], prev + local_dmin)
            adj[i] = new_x - x_sort[i]

        return adj

    # The algorithm in Li, Wang solves the problem given, but there's one tweak to the problem
    # specification that we want to make. Specifically, we'd like to add that, if there's no
    # need to move a particular value, we shouldn't.
    # We do this by simply doing adjustments rightward and leftward and seeing if any values
    # are zero in both. These values stay zero in the final adjustments
    adj1 = adjust_rightward(x_sort)
    adj2 = -adjust_rightward(-x_sort[::-1])[::-1]

    no_shift_required = (adj1 == 0) & (adj2 == 0)

    # Li, Wang now moves everything left by half the max movement
    # this produces the minimum maximum shift
    # alternatively, we subtract by the median, which tries to minimize the sum of absolute value of shifts
    # subtracting by mean tries to minimize the sum of squares of shifts

    adj1 -= np.max(adj1) / 2
    # adj1 -= np.median(adj1)
    # adj1 -= np.mean(adj1)

    adj1 = np.where(no_shift_required, 0, adj1)

    # now adjust, going back to indices of original matrix
    return x + adj1[np.argsort(sort_inds)]


def get_handles_labels(ax):
    """Gets a tuple (handles, labels). Accounts for Seaborn's use of phantom lines, so the lines
    will actually have data on the screen."""
    all_handles, labels = ax.get_legend_handles_labels()
    if len(ax.get_lines()) == 2 * len(all_handles):
        # Seaborn does line plots with a legend by first plotting all of the line data,
        # then making another list that same size of phantom lines without data
        handles = ax.get_lines()[: len(ax.get_lines()) // 2]
    else:
        # none of that weirdness
        handles = all_handles

    return handles, labels


def height(text, ax, **kwargs):
    t = ax.text(0.5, 0.5, text, transform=ax.transAxes, **kwargs)
    ax.figure.canvas.draw()
    bb = t.get_window_extent()
    t.remove()
    return bb.height


def line_labels(ax=None, remove_legend=True):
    """Does automatic line labeling, replacing a legend with side labels."""
    if ax is None:
        ax = plt.gca()

    handles, labels = get_handles_labels(ax)
    labels = ["\n".join(textwrap.wrap(label, width=15)) for label in labels]
    data_to_in = mpl.transforms.CompositeGenericTransform(
        ax.transData, ax.figure.dpi_scale_trans.inverted()
    )

    margin = np.array([height(label, ax) for label in labels])

    y_ends = np.array([handle.get_xydata()[-1] for handle in handles])

    spreaded_y_ends = ax.transData.transform(y_ends)
    spreaded_y_ends[:, 1] = spread(spreaded_y_ends[:, 1], margin)
    spreaded_y_ends = ax.transData.inverted().transform(spreaded_y_ends)

    xmax = ax.transData.inverted().transform(
        (ax.transData.transform((y_ends[:, 0].max(), 0))[0] * 1.05, 0)
    )[0]
    for y_end, spreaded, handle, label in zip(y_ends, spreaded_y_ends, handles, labels):
        color = handle.get_color()
        ax.plot([y_end[0], xmax], [y_end[1], spreaded[1]], ls="--", lw=1, c=color)

        ax.text(xmax, spreaded[1], label, ha="left", va="center", color=color)

    if remove_legend:
        ax.legend().remove()


def ylabel_top(ax=None):
    """Moves the y-axis label to the top, wrapping as necessary."""
    if ax is None:
        ax = plt.gca()

    texts = [
        text
        for text in ax.yaxis.get_children()
        if isinstance(text, mpl.text.Text) and text.get_text() == ax.get_ylabel()
    ]

    text = texts[0]
    text.set_visible(False)

    label_wrap = "\n".join(textwrap.wrap(ax.get_ylabel(), width=15))
    ax.text(
        0,
        1.03,
        label_wrap,
        transform=ax.transAxes,
        ha="right",
        va="baseline",
        color=text.get_color(),
        fontproperties=text.get_font_properties(),
    )



#!/usr/bin/env python3
"""Vega plot themes as config objects."""

import altair as alt

from .colors import LIGHT_SHADES, DARK_SHADES, LIGHT_COLORS, DARK_COLORS
import colorcet as cc

# define the theme by returning the dictionary of configurations
def rho(is_dark: bool):
    empty, lightest, light, medium, dark, darkest = (
        DARK_SHADES if is_dark else LIGHT_SHADES
    )

    def r():
        return {
            "config": {
                "background": empty,
                "view": {"stroke": "transparent", "autosize": "fit",},
                "axis": {
                    "domain": False,
                    "grid": False,
                    "gridColor": light,
                    "tickColor": medium,
                    "gridWidth": 1,
                    "gridOpacity": 0.7,
                    "labelFont": "sans-serif",
                    "labelFontWeight": 300,
                    "labelFontSize": 14,
                    "labelColor": dark,
                    "titleAnchor": "middle",
                    "titleColor": dark,
                    "titleFont": "sans-serif",
                    "titleFontSize": 16,
                    "titleFontWeight": 400,
                    "titlePadding": 10,
                    "zindex": 0,
                },
                "axisBottom": {"domain": True,},
                "axisLeft": {"titleAngle": 0, "titleAlign": "right", "domain": True,},
                "axisQuantitative": {"tickCount": 6,},
                "legend": {
                    "labelColor": dark,
                    "labelFontWeight": 300,
                    "labelFontSize": 14,
                    "titleColor": darkest,
                    "titleFontWeight": 400,
                    "titleFontSize": 16,
                    "titlePadding": 10,
                    "gridLayout": "all",
                    "fillColor": lightest,
                    "padding": 10,
                    "cornerRadius": 6,
                    "rowPadding": 10,
                },
                "title": {
                    "color": darkest,
                    "font": "sans-serif",
                    "fontSize": 24,
                    "fontWeight": 700,
                    "subtitleColor": darkest,
                    "subtitleFont": "sans-serif",
                    "offset": 20,
                    "anchor": "middle",
                },
                "range": {
                    "category": DARK_COLORS,
                    "diverging": cc.coolwarm,
                    "heatmap": cc.CET_L17,
                    "ordinal": {"scheme": "darkmulti" if is_dark else "lightmulti"},
                    "ramp": cc.CET_L17[-30:30:-1] if is_dark else cc.CET_L17[:60:-1],
                },
            }
        }

    return r


RHO_LIGHT = rho(False)()
RHO_DARK = rho(True)()


def setup(is_dark: bool):
    """Sets up Altair according to the given color scheme."""
    alt.themes.register("rho", rho(is_dark))
    alt.themes.enable("rho")
